#!/usr/bin/env python
# -*- coding: utf-8 -*-

#---------------------------------------------------------------------------
#
# SFメモリバイナリイメージエディタ v1.0
# 
# http://sankichi.backdrop.jp/
#---------------------------------------------------------------------------

# Windows10で実行する際はコマンドプロンプト上で
# > chcp 65001
# を実行してから起動すること。

import sys
import glob
import serial
import datetime
import struct
import time

from classes import df_Msg
from classes import dfMenu
from classes import df_Common

# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 0

#---------------------------------------------------------------------------
# 定数など
#---------------------------------------------------------------------------

# ヘルプメッセージ
helpMsg =   'usage: %s SFMimageFile.sfc                   \n'\

# メニュー
menuStr_Top = \
            '\n'\
            '==========================================================\n'\
            '                     <<トップメニュー>>                   \n'\
            '==========================================================\n'\
            'SFメモリカセットイメージファイル: %s\n'\
            '----------------------------------------------------------\n'\
            '1: ディレクトリ情報編集                                   \n'\
            '2: メニュー文字列bitmap編集                               \n'\
            'q: 終了                                                   \n'\
            '----------------------------------------------------------\n'\
            '選択してください: '

#---------------------------------------------------------------------------
#
# メニュー：トップ
#
#---------------------------------------------------------------------------
def menuProc_Top(g_menu):
    # メニュー表示
    print(menuStr_Top % readFileName, end='')
    
    # 入力待ち＆判定
    inStr = input()
    if len(inStr) == 0:
        return g_menu

    elif inStr == 'q':      # 終了
        print('終了します。')
        g_menu.pop()
        return g_menu

    elif inStr == '1':      # メニュー1
        g_menu.append(1)
        return g_menu

    elif inStr == '2':      # メニュー2
        g_menu.append(2)
        return g_menu

    else:                   # 不明コマンド
        print('不明なコマンドです。')
        return g_menu

#---------------------------------------------------------------------------
# メニューループ処理
#---------------------------------------------------------------------------
def menuLoop(sfcBytes):
    g_menu=[0]

    # ループ
    while True:
        if g_menu == [0]:
            # メニュー：トップ
            g_menu = menuProc_Top(g_menu)

        elif g_menu == [0, 1]:
            # メニュー：ディレクトリ情報エディット
            sfcBytes = proc_dirInfo(sfcBytes)
            g_menu.pop()

        elif g_menu == [0, 2]:
            # メニュー：bitmapエディット
            sfcBytes = proc_bitmap(sfcBytes)
            g_menu.pop()

        elif g_menu == []:
            # 終了
            break
    
    # End of while


#---------------------------------------------------------------------------
#
# ディレクトリ情報編集ルーチン
#
#---------------------------------------------------------------------------
def proc_dirInfo(sfcBytes):
    dataChanged = False
    
    # MENU PROGRAMを探す
    if sfcBytes[0x007fc0:(0x007fc0 + 12)] != bytes(
                                                [0x4D, 0x45, 0x4E, 0x55, 
                                                0x20, 0x50, 0x52, 0x4F, 
                                                0x47, 0x52, 0x41, 0x4D]):
        
        print('MENU PROGRAMが存在しないか、検知できませんでした。')
        return sfcBytes
    #print('MENU PROGRAMを検知')

    # ディレクトリ一覧表示
    maxGameNum = printSFMDirectory(sfcBytes)
    #print('%d' % maxGameNum)
    
    # ゲーム番号入力
    gameNumInt = inputGameNum(maxGameNum)
    
    # Gamecode
    print('Gamecodeを編集しますか？y/[n]:', end='')
    inputStr = input()
    if   inputStr == 'y' or inputStr == 'Y':
        # 編集モード
        while True:
            print('Gamecodeを入力してください。(半角0-12文字):', end='')
            str_gameCode = input()
            if len(str_gameCode) > 12:
                #やり直し
                continue
            else:
                #次に進む
                break
        # SFメモリカセットイメージに書き込み
        sfcBytes = editSFMimage_Gamecode(gameNumInt, sfcBytes, str_gameCode)
        dataChanged = True
    
    # タイトル
    print('タイトルを編集しますか？y/[n]:', end='')
    inputStr = input()
    if   inputStr == 'y' or inputStr == 'Y':
        # 編集モード
        while True:
            print('タイトルを入力してください。(sjis変換後44バイト以内):', end='')
            str_title = input()
            titleLen = len(str_title.encode('sjis'))
            if titleLen > 44:
                #やり直し
                print('Err: 変換後データ長さ:%d。もう少し短くしてください。' % titleLen)
                continue
            else:
                #次に進む
                break
        # SFメモリカセットイメージに書き込み
        sfcBytes = editSFMimage_Title(gameNumInt, sfcBytes, str_title)
        dataChanged = True
    
    # Date
    print('Dateを編集しますか？y/[n]:', end='')
    inputStr = input()
    if   inputStr == 'y' or inputStr == 'Y':
        # 編集モード
        while True:
            print('Dateを入力してください。(MM/DD/YYYY or YYYY/MM/DD):', end='')
            str_date = input()
            if len(str_date) != 10:
                #やり直し
                continue
            else:
                #次に進む
                break
        # SFメモリカセットイメージに書き込み
        sfcBytes = editSFMimage_Date(gameNumInt, sfcBytes, str_date)
        dataChanged = True
    
    # Time
    print('Timeを編集しますか？y/[n]:', end='')
    inputStr = input()
    if   inputStr == 'y' or inputStr == 'Y':
        # 編集モード
        while True:
            print('Timeを入力してください。(HH:MM:SS):', end='')
            str_time = input()
            if len(str_time) != 8:
                #やり直し
                continue
            else:
                #次に進む
                break
        # SFメモリカセットイメージに書き込み
        sfcBytes = editSFMimage_Time(gameNumInt, sfcBytes, str_time)
        dataChanged = True
    
    # LAW/NINコード
    print('LAW/NINコードを編集しますか？y/[n]:', end='')
    inputStr = input()
    if   inputStr == 'y' or inputStr == 'Y':
        # 編集モード
        while True:
            print('LAW/NINコードを入力してください。(LAWnnnnn or NINnnnnn):', end='')
            lawninStr = input()
            if len(lawninStr) != 8:
                #やり直し
                continue
            else:
                #次に進む
                break
        # SFメモリカセットイメージに書き込み
        sfcBytes = editSFMimage_LAWNIN(gameNumInt, sfcBytes, lawninStr)
        dataChanged = True
    
    
    # イメージ->ファイル書き込み確認
    if dataChanged == True:
        # もう一度ディレクトリ表示
        printSFMDirectory(sfcBytes)

        # 確認
        while True:
            print('SFメモリカセットイメージファイルの内容を書き換えますか？(y/n):', end='')
            inputStr = input()
            if   inputStr == 'y' or inputStr == 'Y':
                break
            elif inputStr == 'n' or inputStr == 'N':
                return sfcBytes

    # ファイル書き込み
    with open(readFileName, "wb") as fout:
        fout.write(sfcBytes)
    print('SFメモリカセットイメージファイルの内容を書き換えました。')

    return sfcBytes

#---------------------------------------------------------------------------
#
# SFメモリカセットイメージを編集（LAW/NINコード編集）
#
#---------------------------------------------------------------------------
def editSFMimage_LAWNIN(gameNumInt, sfcBytes, strLawNin):
    # SFメモリカセットイメージ始点
    # ディレクトリ情報内の選択ゲーム番号の
    # Gamecodeデータ開始位置を計算
    # この位置から8バイト
    editAddr = 0x060000 + (gameNumInt * 0x002000) + 0x01d1
    
    # SFメモリカセットイメージ書き換え
    return (
        sfcBytes[0:editAddr] + 
        strLawNin.encode('utf-8') + 
        sfcBytes[editAddr+8:]
    )


#---------------------------------------------------------------------------
#
# SFメモリカセットイメージを編集（Time編集）
#
#---------------------------------------------------------------------------
def editSFMimage_Time(gameNumInt, sfcBytes, strTime):
    # SFメモリカセットイメージ始点
    # ディレクトリ情報内の選択ゲーム番号の
    # Gamecodeデータ開始位置を計算
    # この位置から8バイト
    editAddr = 0x060000 + (gameNumInt * 0x002000) + 0x01c9
    
    # SFメモリカセットイメージ書き換え
    return (
        sfcBytes[0:editAddr] + 
        strTime.encode('utf-8') + 
        sfcBytes[editAddr+8:]
    )


#---------------------------------------------------------------------------
#
# SFメモリカセットイメージを編集（Date編集）
#
#---------------------------------------------------------------------------
def editSFMimage_Date(gameNumInt, sfcBytes, strDate):
    # SFメモリカセットイメージ始点
    # ディレクトリ情報内の選択ゲーム番号の
    # Gamecodeデータ開始位置を計算
    # この位置から10バイト
    editAddr = 0x060000 + (gameNumInt * 0x002000) + 0x01bf
    
    # SFメモリカセットイメージ書き換え
    return (
        sfcBytes[0:editAddr] + 
        strDate.encode('utf-8') + 
        sfcBytes[editAddr+10:]
    )


#---------------------------------------------------------------------------
#
# SFメモリカセットイメージを編集（タイトル編集）
#
#---------------------------------------------------------------------------
def editSFMimage_Title(gameNumInt, sfcBytes, str_title):
    # SFメモリカセットイメージ始点
    # ディレクトリ情報内の選択ゲーム番号の
    # Gamecodeデータ開始位置を計算
    # この位置から44バイト
    editAddr = 0x060000 + (gameNumInt * 0x002000) + 0x0013
    
    # SFメモリカセットイメージ書き換え
    return (
        sfcBytes[0:editAddr] + 
        str_title.encode('sjis') + 
        bytes([0x00] * (44 - len(str_title.encode('sjis')))) + 
        sfcBytes[editAddr+44:]
    )


#---------------------------------------------------------------------------
#
# SFメモリカセットイメージを編集（Gamecode編集）
#
#---------------------------------------------------------------------------
def editSFMimage_Gamecode(gameNumInt, sfcBytes, strGamecode):
    # SFメモリカセットイメージ始点
    # ディレクトリ情報内の選択ゲーム番号の
    # Gamecodeデータ開始位置を計算
    # この位置から12バイト
    editAddr = 0x060000 + (gameNumInt * 0x002000) + 0x0007
    
    # 半角スペースパディング
    if len(strGamecode) < 12:
        strGamecode = strGamecode + (' ' * (12-len(strGamecode)))
    
    # SFメモリカセットイメージ書き換え
    return (
        sfcBytes[0:editAddr] + 
        strGamecode.encode('utf-8') + 
        sfcBytes[editAddr+12:]
    )

#---------------------------------------------------------------------------
#
# bitmap編集ルーチン
#
#---------------------------------------------------------------------------
def proc_bitmap(sfcBytes):
    # MENU PROGRAMを探す
    if sfcBytes[0x007fc0:(0x007fc0 + 12)] != bytes(
                                                [0x4D, 0x45, 0x4E, 0x55, 
                                                0x20, 0x50, 0x52, 0x4F, 
                                                0x47, 0x52, 0x41, 0x4D]):
        
        print('MENU PROGRAMが存在しないか、検知できませんでした。')
        return sfcBytes
    #print('MENU PROGRAMを検知')

    # ディレクトリ一覧表示
    maxGameNum = printSFMDirectory(sfcBytes)
    #print('%d' % maxGameNum)
    
    # ゲーム番号入力
    gameNumInt = inputGameNum(maxGameNum)
    
    # bmpファイル名を入力
    bmpFileName = inputFilename('*.bmpファイル名を入力してください。:')

    # bmpファイル読み込み
    (retCode, bmpBytes) = readFileAsBytes(bmpFileName)
    if retCode == -1:
        print('Err: bmpファイル読み込みに失敗しました。')
        return sfcBytes

    # SFメモリカセットイメージを編集（bitmap編集）
    sfcBytes = editSFMimage_bitmap(gameNumInt, sfcBytes, bmpBytes)

    # イメージ->ファイル書き込み確認
    while True:
        print('SFメモリカセットイメージファイルの内容を書き換えますか？(y/n):', end='')
        inputStr = input()
        if   inputStr == 'y' or inputStr == 'Y':
            break
        elif inputStr == 'n' or inputStr == 'N':
            return sfcBytes

    # ファイル書き込み
    with open(readFileName, "wb") as fout:
        fout.write(sfcBytes)
    print('SFメモリカセットイメージファイルの内容を書き換えました。')

    return sfcBytes

#---------------------------------------------------------------------------
#
# SFメモリカセットイメージを編集（bitmap編集）
#
#---------------------------------------------------------------------------
def editSFMimage_bitmap(gameNumInt, sfcBytes, bmpBytes):
    # bmpデータ始点
    # bmpのピクセル情報は左下から右上に向けて格納されているらしい
    # なので以下のアドレスは(0, 12)の位置。
    # bmpファイルには、横に192ピクセル＝192バイト格納された後、
    # (0, 11)の行が192バイト続く
    bmpStartAddr = 0x0436
    
    # bmpデータをy方向に逆にする
    y = 11
    reversed = []
    while y >= 0:
        startIdx = bmpStartAddr + 192*y
        reversed.extend(bmpBytes[startIdx:(startIdx + 192)])
        y = y - 1
    
    # y方向逆にしたデータを基に、8バイトを1バイトにまとめる
    #   ・192x12ピクセルの領域を、24x12が横に8つ並んでいるとみなす
    #   ・1つの24x12のセクションは、48バイトで表現する
    #   ・24バイトが2行並んでいると考える
    #     (1)1行目の偶数byte(0,2,4...) 8bit         [x= 0～ 7]
    #     (2)1行目の奇数byte(1,3,5...) 8bit         [x= 8～15]
    #     (3)2行目の偶数byte(0,2,4...) LSB4bit      [x=16～19]
    #     (4)2行目の奇数byte(1,3,5...) MSB4bit      [x=20～23]
    #     ※(2)のbyteのLSB4bitと、(3)のbyteのMSB4bitはORされる。
    
    byteToBit = []
    x = 0
    while x < 192:
        y = 0
        while y < 12:
            # (1)[x= 0～ 7]　そのまま8バイト->1バイト化
            val = 0
            if reversed[x +  0 + (y * 192)] != 0:
                val = val | 0x80
            if reversed[x +  1 + (y * 192)] != 0:
                val = val | 0x40
            if reversed[x +  2 + (y * 192)] != 0:
                val = val | 0x20
            if reversed[x +  3 + (y * 192)] != 0:
                val = val | 0x10
            if reversed[x +  4 + (y * 192)] != 0:
                val = val | 0x08
            if reversed[x +  5 + (y * 192)] != 0:
                val = val | 0x04
            if reversed[x +  6 + (y * 192)] != 0:
                val = val | 0x02
            if reversed[x +  7 + (y * 192)] != 0:
                val = val | 0x01
            byteToBit.extend(bytes([val]))
            
            # (2)[x= 8～15]　そのまま8バイト->1バイト化
            val = 0
            if reversed[x +  8 + (y * 192)] != 0:
                val = val | 0x80
            if reversed[x +  9 + (y * 192)] != 0:
                val = val | 0x40
            if reversed[x + 10 + (y * 192)] != 0:
                val = val | 0x20
            if reversed[x + 11 + (y * 192)] != 0:
                val = val | 0x10
            if reversed[x + 12 + (y * 192)] != 0:
                val = val | 0x08
            if reversed[x + 13 + (y * 192)] != 0:
                val = val | 0x04
            if reversed[x + 14 + (y * 192)] != 0:
                val = val | 0x02
            if reversed[x + 15 + (y * 192)] != 0:
                val = val | 0x01
            byteToBit.extend(bytes([val]))
            
            y = y + 1
        #   (1)(2)の処理終了後、次の位置へ
        x = x + 12

        y = 0
        while y < 12:
            # (3)2行目の偶数byte(0,2,4...) LSB4bit      [x=16～19]
            val = 0
            if reversed[x +  0 + (y * 192)] != 0:
                val = val | 0x80
            if reversed[x +  1 + (y * 192)] != 0:
                val = val | 0x40
            if reversed[x +  2 + (y * 192)] != 0:
                val = val | 0x20
            if reversed[x +  3 + (y * 192)] != 0:
                val = val | 0x10
            if reversed[x +  4 + (y * 192)] != 0:
                val = val | 0x08
            if reversed[x +  5 + (y * 192)] != 0:
                val = val | 0x04
            if reversed[x +  6 + (y * 192)] != 0:
                val = val | 0x02
            if reversed[x +  7 + (y * 192)] != 0:
                val = val | 0x01
            byteToBit.extend(bytes([val]))
            
            # (4)2行目の奇数byte(1,3,5...) MSB4bit      [x=20～23]
            val = 0
            if reversed[x +  8 + (y * 192)] != 0:
                val = val | 0x80
            if reversed[x +  9 + (y * 192)] != 0:
                val = val | 0x40
            if reversed[x + 10 + (y * 192)] != 0:
                val = val | 0x20
            if reversed[x + 11 + (y * 192)] != 0:
                val = val | 0x10
            byteToBit.extend(bytes([val]))
            
            y = y + 1
        #   (3)(4)の処理終了後、次の位置へ
        x = x + 12
    
    ## test
    ## この時点で、byteToBitは384バイト(0x180バイト)あるはず
    #dump(byteToBit)
    
    # SFメモリカセットイメージ始点
    # ディレクトリ情報内の選択ゲーム番号の
    # bitmapデータ開始位置を計算
    # この位置から384バイト
    editAddr = 0x060000 + (gameNumInt * 0x002000) + 0x003f
    #print('editAddr: %06x' % editAddr)
    
    # SFメモリカセットイメージ書き換え
    return sfcBytes[0:editAddr] + bytes(byteToBit) + sfcBytes[editAddr+384:]

#---------------------------------------------------------------------------
#
# byte配列をダンプする
#
#---------------------------------------------------------------------------
def dump(bytes, addr = 0):
    memory_address = 0 + addr
    ascii_string = ""
    print("Offset   00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F ASCII")
    print("-------- ----------------------------------------------- ----------------")
    
    for byte in bytes:
        ascii_string = ascii_string + chr(validate_byte_as_printable(byte))
        if memory_address%16 == 0:
            print(format(memory_address, '08X'),end = '')
            print(" " + hex(byte)[2:].zfill(2), end='')
        elif memory_address%16 == 15:
            print(" " + hex(byte)[2:].zfill(2),end='')
            print(" " + ascii_string)
            ascii_string = ""
        else:
            print(" " + hex(byte)[2:].zfill(2), end='')
        memory_address = memory_address + 1

def validate_byte_as_printable(byte):
    ## ascii 文字列化をチェック. asciiでなければ '.' を返す ##
    if is_character_printable(byte):
        return byte
    else:
        return 46

def is_character_printable(s):
    ## asciiの英数字・記号の文字列かどうかを判別し、真偽値を返します。
    #if s < 126 and s >= 33:
    if s < 126 and s >= 0x20:
        return True 

#---------------------------------------------------------------------------
#
# バイナリファイルを読み込む
#
#---------------------------------------------------------------------------
def readFileAsBytes(fileName):
    with open(fileName, "rb") as fIn:
        rStr = fIn.read()
        rBytes = bytes(rStr)
        if len(rBytes) > 0:
            print('ファイル%sを読み込みました。' % fileName)
            return 0, rBytes
        else:
            print('Err: ファイル%s: ファイルサイズが不正です。' % fileName)
            return -1, []

#---------------------------------------------------------------------------
# ファイル名をユーザに入力させる
#   不正な値(長さ0)の場合、繰り返し
#   戻り値：入力した値
#
#   ※今後、存在チェックを入れた方がいいかも？
#---------------------------------------------------------------------------
def inputFilename(msg):
    while True:
        print(msg, end='')
        inputStr = input()
        if len(inputStr) == 0:
            continue
        break

    return inputStr

#---------------------------------------------------------------------------
#
# ゲーム番号を入力させる
#
#---------------------------------------------------------------------------
def inputGameNum(maxGameNum):
    gameNumInt = 0
    
    while True:
        print('ゲーム番号を入力してください。(#欄の数値、0-7): ', end='')
        gameNumStr = input()
        if len(gameNumStr) == 0:
            continue
        
        gameNumInt = int(gameNumStr, 0)
        if gameNumInt < 0 or gameNumInt > (maxGameNum - 1):
            continue
        break
    
    return gameNumInt

#---------------------------------------------------------------------------
#
# ディレクトリ情報を画面表示
#
#---------------------------------------------------------------------------
def printSFMDirectory(sfcBytes):
    maxGameNum = 0
    
    print('#  F  B Gamecode     タイトル                                     Date       Time     LAW/NIN ')
    print('- -- -- ------------ -------------------------------------------- ---------- -------- --------')

    i = 0
    addr = 0x060000
    while i < 8:
        # ディレクトリ位置0の情報を持ってくる
        #(retCode, retAry) = df_Common.readSizeByType(s, 0x07, addr, 512)
        retAry = sfcBytes[addr:(addr+512)]
        
        if retAry[0] >= 8:
            # ディレクトリ情報はもう無い
            maxGameNum = i
            break
        
        # 各種情報の計算
        dirIdx = retAry[0]
        numOfFlashBlock = (retAry[4] * 0x100 + retAry[3]) / 4
        numOfSRAMBlock  = (retAry[6] * 0x100 + retAry[5]) / 16
        gamecode = str(bytes(retAry[7:19]), encoding='utf-8', errors='replace')
        sjisName = str(bytes(retAry[0x13:0x3f]), encoding='sjis', errors='replace')
        dateStr  = str(bytes(retAry[0x1bf:0x1c9]), encoding='utf-8', errors='replace')
        timeStr  = str(bytes(retAry[0x1c9:0x1d1]), encoding='utf-8', errors='replace')
        lawnin   = str(bytes(retAry[0x1d1:0x1d9]), encoding='utf-8', errors='replace')
        
        print('%1d %2d %2d %12s %s %s %s %s' % 
            (dirIdx, numOfFlashBlock, numOfSRAMBlock, gamecode, sjisName,
             dateStr, timeStr, lawnin))
        
        # 次のディレクトリ情報位置へ
        i = i + 1
        addr = addr + 0x2000

    return maxGameNum

#---------------------------------------------------------------------------
#
# mainルーチン
#
#---------------------------------------------------------------------------
if __name__ == '__main__':
    # 引数の解析
    args = sys.argv
    
    if len(args) < (1 + 1):
        print(helpMsg % (args[0]))
        sys.exit(0)
    
    readFileName = args[1]
    
    # ファイル読み込み
    (retCode, sfcBytes) = readFileAsBytes(readFileName)
    if retCode == -1:
        print('Err: ファイル読み込みに失敗しました。')
        sys.exit(-1)

    # メニュー処理
    menuLoop(sfcBytes)

