#!/usr/bin/env python
# -*- coding: utf-8 -*-

from classes import df_Common
import struct
import datetime
import time     # sleep
import sys      # exit

# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 1

# メニュー：マークIII／ゲームギア
menuStr_SMSGG = \
            '\n'\
            '==========================================================\n'\
            '           <<メニュー：マークIII／ゲームギア>>            \n'\
            '==========================================================\n'\
            '接続中カセット種別: %s                                    \n'\
            '----------------------------------------------------------\n'\
            '1: 先頭512バイトをダンプ表示                              \n'\
            '2: 指定アドレスから指定サイズダンプ表示                   \n'\
            '3: 領域1(0x4000-0x7fff)切り替え                           \n'\
            '4: 領域2(0x8000-0xbfff)切り替え                           \n'\
            '5: ROMバンク00～ffの先頭16バイト一覧表示                  \n'\
            '6: 16KbyteROM読み出し＆ROMファイル作成                    \n'\
            '7: 32KbyteROM読み出し＆ROMファイル作成                    \n'\
            '8: 64KbyteROM読み出し＆ROMファイル作成                    \n'\
            '9: 指定メガビットROM読み出し＆ROMファイル作成             \n'\
            'b: 戻る                                                   \n'\
            'q: 終了                                                   \n'\
            '----------------------------------------------------------\n'\
            '選択してください: '



#---------------------------------------------------------------------------
#
# メニュー：マークIII／ゲームギア
#
#---------------------------------------------------------------------------
def menuProc(s, stateStr, connected, g_menu):
    # メニュー表示
    print(menuStr_SMSGG % stateStr, end='')
    
    # 入力待ち＆判定
    inStr = input()
    if len(inStr) == 0:
        return g_menu

    elif inStr == 'b':      # 戻る
        g_menu.pop()
        return g_menu

    elif inStr == 'q':      # 終了
        g_menu = []
        return g_menu

    elif inStr == '1':      # 先頭512バイトダンプ
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            print('先頭ダンプ')
            
            """
            iType = 0x27
            while iType < 256:
                print('read0x%02x...' % (iType))
                (retCode, retAry) = df_Common.readSizeByType(s, iType, 0, 512)
                df_Common.dump(retAry)
                print('read0x%02x retCode: %d' % (iType, retCode))
                iType = iType + 1
                time.sleep(1.0)
            """
            (retCode, retAry) = df_Common.readSizeByType(s, 0x2b, 0, 512)
            #print('retCode: %d' % retCode)
            df_Common.dump(retAry)
            print('')
        return g_menu

    elif inStr == '2':      # 指定アドレス指定サイズダンプ
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('指定アドレス指定サイズダンプ')

            # アドレスを入力させる
            stAddr = df_Common.inputHexValue('開始アドレス(16進)を入力してください。: ')
            
            # サイズを入力させる
            stSize = df_Common.inputHexValue('サイズ(16進)を入力してください。: ')

            (retCode, retAry) = df_Common.readSizeByType(s, 0x2b, stAddr, stSize)
            #print('retCode: %d' % retCode)
            df_Common.dump(retAry, stAddr)
            print('')
        return g_menu

    elif inStr == '3':      # 領域1(0x4000-0x7fff)切り替え
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('領域1(0x4000-0x7fff)切り替え')

            # バンクNoを入力させる
            while True:
                bankNo = df_Common.inputHexValue('バンク番号(16進,0～ff)を入力してください。: ')
                if bankNo >= 0x00 and bankNo <= 0xff:
                    break

            # 領域1(0x4000-0x7fff)切り替え
            setBankArea1(s, bankNo)
            print('完了。')
            print('')
        return g_menu

    elif inStr == '4':      # 領域2(0x8000-0xbfff)切り替え
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('領域2(0x8000-0xbfff)切り替え')

            # バンクNoを入力させる
            while True:
                bankNo = df_Common.inputHexValue('バンク番号(16進,0～ff)を入力してください。: ')
                if bankNo >= 0x00 and bankNo <= 0xff:
                    break

            # 領域2(0x8000-0xbfff)切り替え
            setBankArea2(s, bankNo)
            print('完了。')
            print('')
        return g_menu

    elif inStr == '5':      # ROMバンク00～ffの先頭16バイト一覧表示
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('ROMバンク00～ffの先頭16バイト一覧表示')

            print("Bank 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F ASCII")
            print("---- ----------------------------------------------- ----------------")

            bankNo = 0x00
            while bankNo < 0x100:
                # ダンプ用
                memory_address = 0
                ascii_string = ""
                # 領域2(0x8000-0xbfff)切り替え
                setBankArea2(s, bankNo)
                # 読み
                (retCode, bytes) = df_Common.readSizeByType(s, 0x2b, 0x8000, 16)
                # 表示
                for byte in bytes:
                    ascii_string = ascii_string + chr(df_Common.validate_byte_as_printable(byte))
                    if memory_address%16 == 0:
                        print(format(bankNo, '04X'),end = '')
                        print(" " + hex(byte)[2:].zfill(2), end='')
                    elif memory_address%16 == 15:
                        print(" " + hex(byte)[2:].zfill(2),end='')
                        print(" " + ascii_string)
                        ascii_string = ""
                    else:
                        print(" " + hex(byte)[2:].zfill(2), end='')
                    memory_address = memory_address + 1
                # バンクNo増加
                bankNo = bankNo + 1

            print('完了。')
            print('')
        return g_menu

    elif inStr == '6':      # 16KByteファイル化
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('16KByteファイル化')

            # ファイル名入力
            fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')
            
            fileAry = []
            addr = 0
            dumpSize = 1024
            while addr < 1024*16:
                # ファイル出力用配列にバンバン追加
                print('\rアドレス {:08x} から読み出し中...'.format(addr), end='')
                (retCode, retAry) = df_Common.readSizeByType(s, 0x2b, addr, dumpSize)
                if retCode == 0:
                    fileAry.extend(retAry)
                    addr = addr + dumpSize
                else:
                    print('リトライ実施(retCode: %d)' % retCode)
                    #s.rtscts = 1
                    continue
            print('完了。')
            
            # 配列をファイル出力
            with open(fileNameStr, "wb") as fOut:
                for x in fileAry:
                    fOut.write(struct.pack("B", x))
            print('ファイル: %sを出力しました。' % fileNameStr)
            print('')
        return g_menu


    elif inStr == '7':      # 32KByteファイル化
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('32KByteファイル化')

            # ファイル名入力
            fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')
            
            fileAry = []
            addr = 0
            dumpSize = 1024
            while addr < 1024*32:
                # ファイル出力用配列にバンバン追加
                print('\rアドレス {:08x} から読み出し中...'.format(addr), end='')
                (retCode, retAry) = df_Common.readSizeByType(s, 0x2b, addr, dumpSize)
                if retCode == 0:
                    fileAry.extend(retAry)
                    addr = addr + dumpSize
                else:
                    print('リトライ実施(retCode: %d)' % retCode)
                    #s.rtscts = 1
                    continue
            print('完了。')
            
            # 配列をファイル出力
            with open(fileNameStr, "wb") as fOut:
                for x in fileAry:
                    fOut.write(struct.pack("B", x))
            print('ファイル: %sを出力しました。' % fileNameStr)
            print('')
        return g_menu

    elif inStr == '8':      # 64KByteファイル化
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('64KByteファイル化')

            # ファイル名入力
            fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')
            
            fileAry = []
            addr = 0
            dumpSize = 1024
            while addr < 1024*64:
                # ファイル出力用配列にバンバン追加
                print('\rアドレス {:08x} から読み出し中...'.format(addr), end='')
                (retCode, retAry) = df_Common.readSizeByType(s, 0x2b, addr, dumpSize)
                if retCode == 0:
                    fileAry.extend(retAry)
                    addr = addr + dumpSize
                else:
                    print('リトライ実施(retCode: %d)' % retCode)
                    #s.rtscts = 1
                    continue
            print('完了。')
            
            # 配列をファイル出力
            with open(fileNameStr, "wb") as fOut:
                for x in fileAry:
                    fOut.write(struct.pack("B", x))
            print('ファイル: %sを出力しました。' % fileNameStr)
            print('')
        return g_menu

    elif inStr == '9':      # 指定メガビットROM読み出し＆ROMファイル作成
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('指定メガビットROM読み出し＆ROMファイル作成')

            # ファイル名入力
            fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')
            
            # サイズ入力
            while True:
                mBit = df_Common.inputValue('カセットのメガビット数を入力してください。(1～32): ')
                if mBit < 1 or mBit > 32:
                    continue
                else:
                    break
            # 最大バンク数計算
            #  1Mbit =   8bank
            #  2Mbit =  16bank
            #  4Mbit =  32bank
            # ・・・
            # 32Mbit = 256bank
            maxBank = mBit * 8
            #print("バンク%02xまで読み出します。" % (maxBank))
            
            # 読み出し
            fileAry = []
            bankNo = 0
            while bankNo < maxBank:
                # ファイル出力用配列にバンバン追加
                print('\rバンク {:02x} から読み出し中...'.format(bankNo), end='')
                # バンク切り替え
                # 領域2(0x8000-0xbfff)切り替え
                setBankArea2(s, bankNo)

                addr = 0x8000
                dumpSize = 1024
                while addr < 0xc000:
                    #print('\rアドレス {:04x} から読み出し中...'.format(addr), end='')
                    (retCode, retAry) = df_Common.readSizeByType(s, 0x2b, addr, dumpSize)
                    if retCode == 0:
                        fileAry.extend(retAry)
                        addr = addr + dumpSize
                    else:
                        print('リトライ実施(retCode: %d)' % retCode)
                        continue
                # バンク増加
                bankNo = bankNo + 1
            
            print('完了。')
            
            # 配列をファイル出力
            with open(fileNameStr, "wb") as fOut:
                for x in fileAry:
                    fOut.write(struct.pack("B", x))
            print('ファイル: %sを出力しました。' % fileNameStr)
            print('')
        return g_menu

    else:                   # 不明コマンド
        print('不明なコマンドです。')
        return g_menu


#---------------------------------------------------------------------------
#
# 領域1(0x4000-0x7fff)切り替え
#
#---------------------------------------------------------------------------
def setBankArea1(s, bankNo):
    writeBytesGG(s, 0xfffc, bytes([0x80]))
    writeBytesGG(s, 0xfffe, bytes([bankNo]))

#---------------------------------------------------------------------------
#
# 領域2(0x8000-0xbfff)切り替え
#
#---------------------------------------------------------------------------
def setBankArea2(s, bankNo):
    writeBytesGG(s, 0xfffc, bytes([0x80]))
    writeBytesGG(s, 0xffff, bytes([bankNo]))

#---------------------------------------------------------------------------
# bytes配列を指定アドレスに書き込み
#       戻り値：
#               retCode: 
#                    0:成功
#                   -1:書き込み要求コマンドの応答タイムアウト
#                   -2:書き込み要求コマンドに肯定応答でない応答
#                   -3:書き込みデータ本体受信タイムアウト
#
#       メモ：writeBytes_SFC()でも同じことをやっている。共通化したほうが・・・？
#---------------------------------------------------------------------------
def writeBytesGG(s, addrInt, datas):
    size = len(datas).to_bytes(4, 'little')
    addr = addrInt.to_bytes(4, 'little')

    df_Common.putLog('writeBytesGG(): 書き込みデータ 先頭: 0x%02x' % datas[0])
    
    # 1. 書き込みコマンド
    s.write(bytes([
        0x2c, 0x00, 0x00, 0x00,             # 
        0x08, 0x00, 0x00, 0x00,             # 不明 8
        size[0], size[1], size[2], size[3], # サイズ？
        addr[0], addr[1], addr[2], addr[3], # 開始アドレス
        size[0], size[1], size[2], size[3], # サイズ？
        ]))

    # 2. 書き込みコマンド応答待ち
    respWaitCount = 0
    while True:
        if s.in_waiting > 0:
            df_Common.putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            respWaitCount = respWaitCount + 1
            df_Common.putLog('0.01秒待機。')
            time.sleep(0.01)
            if(respWaitCount > 40):
                print('ERR: 書き込みコマンド応答タイムアウト')
                return -1

    # 3. 書き込みコマンド応答の受信と判定
    readStr = s.read(8)
    readBytes = bytes(readStr)
    df_Common.putLog('書き込みコマンド応答受信。')
    df_Common.debugDump(readBytes)

    if readBytes == bytes([0,0,0,0,0,0,0,0]):
        df_Common.putLog('肯定応答を受信。')
    else:
        df_Common.putLog('肯定応答でない応答を受信。続行。')

    # 4. データ書き込み
    df_Common.putLog("s.write(datas)開始")
    s.write(datas)
    df_Common.putLog("s.write(datas)終了")

    # 3. データ書き込み応答待ち
    respWaitCount = 0
    while True:
        if s.in_waiting > 0:
            df_Common.putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            respWaitCount = respWaitCount + 1
            df_Common.putLog('0.01秒待機。')
            time.sleep(0.01)
            if(respWaitCount > 40):
                print('ERR: データ書き込み応答タイムアウト')
                return -2

    # 4. データ書き込み応答の受信と判定
    readStr = s.read(8)
    readBytes = bytes(readStr)
    df_Common.putLog('データ書き込み応答受信。')
    df_Common.debugDump(readBytes)

    if readBytes == bytes([0,0,0,0,0,0,0,0]):
        df_Common.putLog('肯定応答を受信。')
    else:
        df_Common.putLog('肯定応答でない応答を受信。続行。')

    # ここまで来たら成功とみなす
    return 0

