#!/usr/bin/env python
# -*- coding: utf-8 -*-

import struct
import time     # sleep
from classes import df_Common

# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 0

#---------------------------------------------------------------------------
# bytes配列を指定アドレスに書き込み
#       戻り値：
#               retCode: 
#                    0:成功
#                   -1:書き込み要求コマンドの応答タイムアウト
#                   -2:書き込み要求コマンドに肯定応答でない応答
#                   -3:書き込みデータ本体受信タイムアウト
#
#       メモ：
#---------------------------------------------------------------------------
def writeBytes_SFC(s, addrInt, datas):
    size = len(datas).to_bytes(4, 'little')
    addr = addrInt.to_bytes(4, 'little')

    df_Common.putLog('writeBytes_SFC(): 書き込みデータ 先頭: 0x%02x' % datas[0])
    

    # 1. 書き込みコマンド
    s.write(bytes([
        0x0c, 0x00, 0x00, 0x00,             # 
        0x08, 0x00, 0x00, 0x00,             # 不明 8
        size[0], size[1], size[2], size[3], # サイズ？
        addr[0], addr[1], addr[2], addr[3], # 開始アドレス
        size[0], size[1], size[2], size[3], # サイズ？
        ]))

    # 2. 書き込みコマンド応答待ち
    respWaitCount = 0
    while True:
        if s.in_waiting > 0:
            df_Common.putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            respWaitCount = respWaitCount + 1
            df_Common.putLog('0.01秒待機。')
            time.sleep(0.01)
            if(respWaitCount > 40):
                print('ERR: 書き込みコマンド応答タイムアウト')
                return -1

    # 3. 書き込みコマンド応答の受信と判定
    readStr = s.read(8)
    readBytes = bytes(readStr)
    df_Common.putLog('書き込みコマンド応答受信。')
    df_Common.debugDump(readBytes)

    if readBytes == bytes([0,0,0,0,0,0,0,0]):
        df_Common.putLog('肯定応答を受信。')
    else:
        #df_Common.putLog('肯定応答でない応答を受信。')
        #print('ERR: 2Fコマンドに肯定応答でない応答')
        #return -2, retAry
        df_Common.putLog('肯定応答でない応答を受信。続行。')

    # 4. データ書き込み
    s.write(datas)

    # 3. データ書き込み応答待ち
    respWaitCount = 0
    while True:
        if s.in_waiting > 0:
            df_Common.putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            respWaitCount = respWaitCount + 1
            df_Common.putLog('0.01秒待機。')
            time.sleep(0.01)
            if(respWaitCount > 40):
                print('ERR: データ書き込み応答タイムアウト')
                return -2

    # 4. データ書き込み応答の受信と判定
    readStr = s.read(8)
    readBytes = bytes(readStr)
    df_Common.putLog('データ書き込み応答受信。')
    df_Common.debugDump(readBytes)

    if readBytes == bytes([0,0,0,0,0,0,0,0]):
        df_Common.putLog('肯定応答を受信。')
    else:
        df_Common.putLog('肯定応答でない応答を受信。続行。')

    # ここまで来たら成功とみなす
    return 0

#---------------------------------------------------------------------------
# [SFC] 与えられたアドレスをヘッダとみなした場合、
#       Checksum complement＆ChecksumのXOR結果が0xffffになるか？
#       戻り値：タプル
#               retCode
#                   0 : 成功
#                   -1: 失敗
#                   retCode == 0のとき、以下に有効な値が入る
#                       titleStr
#                       speedVal
#                       mapModeVal
#                       mapModeStr
#                       romSizeVal
#                       ramSizeVal
#                       checkSumVal
#                       checkSumCompVal
#
#       メモ：
#---------------------------------------------------------------------------
def get_SFCHeaderInfo(s):
    # ROM種別(HiROM, LoROM, ExHiROM)判定
    # ヘッダ情報が入っているはずの領域をアクセスし、
    # チェックサム部から判定する
    headerAddr = 0
    while True:
        # LoROMか？
        (retVal, result) = check_SFCHeaderChecksumComplement(s, 0x7fc0)
        if retVal == -1:
            # 終了
            print('Err: 処理中にエラー発生。')
            return -1, '', 0, 0, '', 0, 0
        else:
            if result == True:
                print('接続中のカセットはLoROMです。')
                headerAddr = 0x7fc0
                # breakして処理続行
                break

        # LoROMか？
        (retVal, result) = check_SFCHeaderChecksumComplement(s, 0xffc0)
        if retVal == -1:
            # 終了
            print('Err: 処理中にエラー発生。')
            return -1, '', 0, 0, '', 0, 0
        else:
            if result == True:
                print('接続中のカセットはHiROMです。')
                headerAddr = 0xffc0
                # breakして処理続行
                break

        # ExHiROMか？
        (retVal, result) = check_SFCHeaderChecksumComplement(s, 0x40ffc0)
        if retVal == -1:
            # 終了
            print('Err: 処理中にエラー発生。')
            return -1, '', 0, 0, '', 0, 0
        else:
            if result == True:
                print('接続中のカセットはExHiROMです。')
                headerAddr = 0x40ffc0
                # breakして処理続行
                break

    # この時点でheaderAddr == 0だったら判定に失敗している
    if headerAddr == 0:
        print('Err: 接続中のカセットのヘッダー部を特定できませんでした。')
        return -1, '', 0, 0, '', 0, 0
    
    # ヘッダ情報表示
    (retCode, romDat) = df_Common.readSizeByType(s, 0x07, headerAddr, 32)
    if retCode != 0:
        print('Err: 処理中にエラー発生。')
        return -1, '', 0, 0, '', 0, 0
    else:
        titleStr = str(bytes(romDat[0:0x15]), encoding='utf-8', errors='replace')
        # FFD5h
        #   Bit4   Speed (0=Slow, 1=Fast)              (Slow 200ns, Fast 120ns)
        if romDat[0x15] & 0x10 == 0:
            speedVal = 0
        else:
            speedVal = 1
        #   Bit3-0 Map Mode
        mapModeVal = romDat[0x15] & 0x0f
        mapModeStr = 'Unknown'
        if mapModeVal == 0:
            mapModeStr = 'LoROM/32K Banks'
        elif mapModeVal == 1:
            mapModeStr = 'HiROM/64K Banks'
        elif mapModeVal == 2:
            mapModeStr = 'LoROM/32K Banks + S-DD1'
        elif mapModeVal == 3:
            mapModeStr = 'LoROM/32K Banks + SA-1'
        elif mapModeVal == 5:
            mapModeStr = 'HiROM/64K Banks(ExHiROM)'
        elif mapModeVal == 0x0A:
            mapModeStr = 'HiROM/64K Banks + SPC7110'

        romSizeVal = 1 << romDat[0x17]
        ramSizeVal = 1 << romDat[0x18]

        checkSumCompVal = romDat[0x1c] + (romDat[0x1d] * 0x100)
        checkSumVal     = romDat[0x1e] + (romDat[0x1f] * 0x100)

    return 0, titleStr, speedVal, mapModeVal, mapModeStr, romSizeVal, ramSizeVal, checkSumVal, checkSumCompVal


#---------------------------------------------------------------------------
# [SFC] 与えられたアドレスをヘッダとみなした場合、
#       Checksum complement＆ChecksumのXOR結果が0xffffになるか？
#       戻り値：2つの値のタプル
#               retCode
#                   0 : 成功
#                   -1: 失敗
#               checkResult
#                    True :(retCode==0のとき)0xffffになる
#                    False:(retCode==0のとき)        ならない
#
#       メモ：
#---------------------------------------------------------------------------
def check_SFCHeaderChecksumComplement(s, addr):
    targetAddr = addr + 0x1c  # 0xffc0が来た時に0xffdcを参照
    (retCode, romDat) = df_Common.readSizeByType(s, 0x07, targetAddr, 4)
    
    if retCode != 0:
        putLog('check_SFCHeaderChecksumComplement() Err: データ読み込み失敗。')
        return -1, False
    else:
        if (romDat[0] * 256 + romDat[1]) ^ (romDat[2] * 256 + romDat[3]):
            return 0,  True
        else:
            return 0,  False

#---------------------------------------------------------------------------
# [SFC] LoROMのROM上の実アドレスを計算して返す
#
#       メモ：
#---------------------------------------------------------------------------
def calcRealAddrSFC_LoROM(addr):
    # 0x000000 ～ 0x3effffまで
    if   addr >= 0x000000 and addr <= 0x3effff:
        return (addr & 0x007fff) + (addr >> 15) * 0x10000 + 0x8000
    # 0x3f0000 ～
    else:
        return (addr & 0x007fff) + (addr >> 15) * 0x10000 + 0x8000 + 0x800000

#---------------------------------------------------------------------------
# [SFC] HiROMのROM上の実アドレスを計算して返す
#
#       メモ：
#---------------------------------------------------------------------------
def calcRealAddrSFC_HiROM(addr):
    if addr >= 0x000000 and addr < 0x3e0000:
        return addr + 0x400000
    else:
        return addr + (0xfe0000 - 0x3e0000)

        
#---------------------------------------------------------------------------
# [SFC] ROM配列のチェックサムを計算する
#
#       メモ：
#         単純なバイト和らしい？？？？
#---------------------------------------------------------------------------
def calcCheckSum_SFC(romAry):
    countMax = len(romAry)

    i = 0
    sum = 0

    while i < countMax:
        sum = sum + romAry[i]
        i = i + 1
    return (sum & 0xffff)


